# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_Smoothing.ipynb (unless otherwise specified).

__all__ = ['__single', '__double', 'holt_winters', '__smoothing', 'SINGLE', 'DOUBLE']

# Cell
import numpy as np
from numba import jit
import pandas as pd

from time_series_model_basics import simulate_data

# Cell
def __single(
    ts,
    alpha=0.1,
):

    if alpha > 1 or alpha < 0:
        raise Exception("Alpha should be on the interval (0,1)")

    n = len(ts)
    f = np.full((n,), np.nan)

    f[1] = ts[0]

    for i in range(2, n):
        f[i] = alpha * ts[i - 1] + (1 - alpha) * f[i - 1]

    return f

# Cell
@jit(nopython=True)
def __double(
    ts,
    alpha=0.1,
    beta=0.1,
    s_0=None,
    b_0=None,
    m=1,
):

    if alpha > 1 or alpha < 0:
        raise Exception("alpha should be on the interval (0,1)")
    if beta > 1 or beta < 0:
        raise Exception("beta should be on the interval (0,1)")

    n = len(ts) + m - 1
    s = np.full((n,), np.nan)
    b = np.full((n,), np.nan)
    f = np.full((n,), np.nan)
    tss = np.append(ts, np.full(m, np.nan))

    s[0] = tss[0] if s_0 is None else s_0
    b[0] = tss[1] - tss[0] if b_0 is None else b_0

    for i in range(1, n):

        s[i] = alpha * tss[i] + (1 - alpha) * (s[i - 1] + b[i - 1])
        b[i] = beta * (s[i] - s[i - 1]) + (1 - beta) * b[i - 1]

        f[i - 1 + m] = s[i - 1] + m * b[i - 1]

    return f

# Cell
@jit(nopython=True)
def holt_winters(
    ts,
    l=1,
    alpha=0.1,
    beta=0.1,
    gamma=0.1,
    s_0=None,
    b_0=None,
    m=1,
):

    if alpha > 1 or alpha < 0:
        raise Exception("alpha should be on the interval (0,1)")
    if beta > 1 or beta < 0:
        raise Exception("beta should be on the interval (0,1)")
    if gamma > 1 or gamma < 0:
        raise Exception("gamma should be on the interval (0,1)")

    if (l <= 0) and (l - int(l) != 0):
        raise Exception("l should be a Natural Number")

    n = len(ts) + m
    #complete cycles
    N = (n - m) // l

    if N == 0:
        raise Exception("it should be at least one complete cycle")

    s = np.full((n,), np.nan)
    b = np.full((n,), np.nan)
    f = np.full((n,), np.nan)
    c = np.full((n,), np.nan)
    A = np.full((N,), np.nan)

    tss = np.append(ts, np.full(m, np.nan))

    s[0] = tss[0] if s_0 is None else s_0

    b[0] = (1 / l**2) * np.sum(tss[l:2 * l] - tss[:l])

    for j in range(1, N + 1):
        A[j - 1] = (1 / l) * np.sum(tss[l * (j - 1):l * (j - 1) + l])

    for i in range(l):
        c[i] = (1 / N) * np.sum(tss[i:l * (N - 1) + i + 1:l] / A)

    for i in range(1, n):
        s[i] = alpha * (tss[i] - c[i - l]) + (1 - alpha) * (s[i - 1] + b[i - 1])
        b[i] = beta * (s[i] - s[i - 1]) + (1 - beta) * b[i - 1]
        c[i] = gamma * (tss[i] - s[i - 1] - b[i - 1]) + (1 - gamma) * c[i - l]

        f[i - 1 + m] = s[i - i] + m * b[i] + c[i - l + 1 + ((m - 1) % l)]

    return f

# Cell
def __smoothing(
    *args,
    kind=None,
    df=None,
    ts_col=None,
    **kwargs,
):
    """
    Forescasts  with Exponential Smoothing

    -----
    Parameters
    -----

    kind : str , default 'simple'. It can be:

         'simple' : Single Smoothing.
         'double' : Double Smoothing.
    df : DataFrame, default None. If df is None a dataframe
         with simulated data is generated.

    ts_col : str, default None. The column name with the
         time series. Set to 'time_series' when df is None.
    *args : int.  Each value represent a Moving
           Average Forecast and its corresponding
           window size
    **kwargs : passed to __pandas_time_series

    -------
    Returns
    -------

    tuple: First element is the Pandas Dataframe and the second
           is ploty's figure object
    """

    if len(args) == 0:
        raise Exception("list in *args should include at least one value")

    if kind not in ('single', 'double'):
        raise Exception("Only supported methods are: 'simple', 'weighted'")

    if df is None:
        df = simulate_data.pandas_time_series(**kwargs)
        ts_col = 'time_series'

    if ts_col is None:
        raise Exception('When df is not None, ts_col should be specified')

    ts = df[ts_col].to_numpy()
    cols = [ts_col]
    for a in args:
        col = f'{kind}_{a}'
        cols.append(col)
        if kind == 'single':
            df[col] = __single(
                ts=ts,
                alpha=a,
            )
        elif kind == 'double':
            df[col] = __double(
                ts=ts,
                alpha=a[0],
                beta=a[1],
            )

    fig = df[cols].plot(
        backend='plotly',
        title=f'Smoothing {kind}',
    )

    fig.update_layout(template="plotly_dark",)

    return df, fig

# Cell
def SINGLE(
    *args,
    df=None,
    ts_col=None,
    **kwargs,
):
    """
    Forescasts  with Single Smoothing

    -----
    Parameters
    -----
    df : DataFrame, default None. If df is None a dataframe
         with simulated data is generated.
    ts_col : str, default 'time_series'. The column name with the
         time series. Ignored if df is None.
    *args : int.  Each value represent a Single Smoothing
         parameter
    **kwargs : passed to __pandas_time_series
    -------
    Returns
    -------
    tuple: First element is the Pandas Dataframe and the second
           is ploty's fig object
    """

    return __smoothing(
        *args,
        kind='single',
        df=df,
        ts_col=ts_col,
        **kwargs,
    )

# Cell
def DOUBLE(*args, df=None, ts_col=None, **kwargs):
    """
    Forescasts  with Double Smoothing

    -----
    Parameters
    -----
    df : DataFrame, default None. If df is None a dataframe
         with simulated data is generated.
    ts_col : str, default 'time_series'. The column name with the
         time series. Ignored if df is None.
    *args : Each pair reprents the two parameters
         of the double smoothing
    **kwargs : passed to __pandas_time_series
    -------
    Returns
    -------
    tuple: First element is the Pandas Dataframe and the second
           is ploty's fig object
    """

    return __smoothing(
        *args,
        kind='double',
        df=df,
        ts_col=ts_col,
        **kwargs,
    )


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_Metrics.ipynb (unless otherwise specified).

__all__ = ['__total_error', '__absolute_error', '__squared_error', '__mean_error', '__mean_absolute_error',
           '__mean_squared_error', '__root_mean_square_error', '__mean_percentage_error',
           '__mean_absolute_percent_error', 'SUMMARY']

# Cell
import numpy as np
from numba import jit
import pandas as pd
from time_series_model_basics import simulate_data, moving_average, metrics
from typing import List

# Cell
def __total_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the total error:

    $\sum^{n}_{i}(ts[i] - f[i]), n = len(ts) = len(f)$

    .Ignores nan values in times-series or the forecast.

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """

    d = ts - f
    return np.sum(d, where=~np.isnan(d))

# Cell
def __absolute_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the absolute error:

    $\sum^{n}_{i}|ts[i] - f[i]|, n = len(ts) = len(f)$

    .Ignores nan values in times-series or the forecast.

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """

    d = np.abs(ts - f)

    return np.sum(d, where=~np.isnan(d))

# Cell
def __squared_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the squared error:

    $\sum^{n}_{i}|ts[i] - f[i]|**2, n = len(ts) = len(f)$

    .Ignores nan values in times-series or the forecast.

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """

    d2 = (ts - f)**2
    return np.sum(d2, where=~np.isnan(d2))

# Cell
def __mean_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the mean error:

    $(1/n)\sum^{n*}_{i}ts[i] - f[i], n* = len(ts) = len(f)$

    .Ignores nan values in times-series or the forecast.
     Value n is n* minus the ignored values

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """
    d = ts - f
    w = ~np.isnan(d)
    n = len(d[w])

    return np.divide(
        np.sum(d, where=w),
        n,
        where=n > 0,
    )

# Cell
def __mean_absolute_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the mean absolute error:

    $$(1/n)\sum^{n}_{i} | ts[i] - f[i]|, n = len(ts) = len(f)$$

    .Ignores nan values in times-series or the forecast.

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """

    d = np.abs(ts - f)
    w = ~np.isnan(d)
    n = len(d[w])

    return np.divide(
        np.sum(d, where=w),
        n,
        where=n > 0,
    )

# Cell
def __mean_squared_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the mean squared error:

    $$(1/n)\sum^{n}_{i} | ts[i] - f[i]|**2, n = len(ts) = len(f)$$

    .Ignores nan values in times-series or the forecast.

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """
    d = (ts - f)**2
    w = ~np.isnan(d)
    n = len(d[w])

    return np.divide(
        np.sum(d, where=w),
        n,
        where=n > 0,
    )

# Cell
def __root_mean_square_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the root mean square error:

    $$\sqrt{(1/n)\sum^{n}_{i} | ts[i] - f[i]|**2},  n = len(ts) = len(f)$$

    .Ignores nan values in times-series or the forecast.

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """

    return np.sqrt(__mean_squared_error(ts=ts, f=f))

# Cell
def __mean_percentage_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the root mean percentage error:

    $$ (1/n)\sum ( ts[i] - f[i]) /ts[i],  n* = len(ts) = len(f)$$

    .Ignores nan values and division by zero.
    n is n* minus the ignored values

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """

    pe = np.divide(ts - f, ts, where=ts != 0)
    w = ~np.isnan(pe)

    n = len(pe[w])

    return np.divide(
        np.sum(pe, where=w),
        n,
        where=n > 0,
    )

# Cell
def __mean_absolute_percent_error(
    *,
    ts: np.array,
    f: np.array,
) -> float:
    """
    Computes the mean absolute percentage error:

    $$ (1/n)\sum | ts[i] - f[i]  /ts[i] \vert,  n* = len(ts) = len(f)$$

    .Ignores nan values and division by zero.
    n is n* minus the ignored values

    ------
    Parameters
    ------

    ts : np.array with the time-series
    f :  np.array with the forecast

    -------
    Returns
    -------

    """

    pe = abs(np.divide(ts - f, ts, where=ts != 0))
    w = ~np.isnan(pe)

    n = len(pe[w])

    return np.divide(
        np.sum(pe, where=w),
        n,
        where=n > 0,
    )

# Cell
def SUMMARY(
    df: pd.DataFrame = None,
    *,
    val_col: str = None,
    pred_cols: List[str] = None,
) -> pd.DataFrame:
    """
    Summary of Prediction Metrics

    -----
    Parameters
    -----

    df : dataframe , dafault None. If None it will generate a simulated dataframe.
    val_col : str , default None.  Name of the column with the actual values.
              It should be provided when a datafre is provided.
    pred_cols : List[str] . Names of the columns with the predictions for the values
              If not provided it will take all dataframe columns except val_col

    -----
    Returns
    ------

    dataframe : Summary results


    """
    if df is None:
        df = simulate_data.pandas_time_series()
        df, _ = moving_average.SMA(
            1,
            4,
            df=df,
        )

        val_col = 'time_series'

    if val_col is None:
        raise Exception('When df is provided val_col cannot be None')

    if pred_cols is None:
        pred_cols = [x for x in df.columns if x != val_col]

    actual_values = df[val_col].to_numpy()

    metric_names = [fn for fn in metrics.__all__ if fn[:2] == '__']
    result_df = pd.DataFrame()

    for col in pred_cols:
        forecast = df[col].to_numpy()
        result_df[col] = pd.Series({
            ' '.join(fn[2:].split('_')[:-1]): eval(fn)(ts=actual_values,
                                                       f=forecast)
            for fn in metric_names
        })
    result_df.index.name = 'Error'
    return result_df


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_SimulateData.ipynb (unless otherwise specified).

__all__ = ['__numpy_time_series', 'pandas_time_series']

# Cell
import numpy as np
import pandas as pd

# Cell
def __numpy_time_series(
    num_points=30,
    level=3,
    trend_amplitude=1 / 3,
    sinus_amplitude=1,
    noise_amplitude=2 / 1,
    start=0,
    finish=4 * np.pi,
):

    interval = np.linspace(
        start,
        finish,
        num=num_points + 1,
    )

    season = np.sin(interval[:-1])
    trend = np.vectorize(lambda x: trend_amplitude * x)(interval[:-1])
    noise = noise_amplitude * np.random.random((num_points,))
    signal = level + season + trend + noise
    return interval, signal

# Cell
def pandas_time_series(ts_col='time_series', **kwargs):
    """
    Generates a Dataframe with simulated data

    -----
    Parameters
    -----
    ts_column : str. Default 'time_series'. Name of time series column
    **kwargs : Argumetns of the __numpy_time-series

    -------
    Returns
    -------

    Dataframe :

    """

    interval, d = __numpy_time_series(**kwargs)

    df = pd.DataFrame(
        {ts_col: np.append(d, np.nan)},
        index=interval,
    )
    df.index = interval
    df.index.name = 'Time Series Units'

    return df


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"__numpy_time_series": "00_SimulateData.ipynb",
         "pandas_time_series": "00_SimulateData.ipynb",
         "simple": "01_MovingAverage.ipynb",
         "weighted": "01_MovingAverage.ipynb",
         "__moving_average": "01_MovingAverage.ipynb",
         "SMA": "01_MovingAverage.ipynb",
         "WMA": "01_MovingAverage.ipynb",
         "__single": "02_Smoothing.ipynb",
         "__double": "02_Smoothing.ipynb",
         "holt_winters": "02_Smoothing.ipynb",
         "__smoothing": "02_Smoothing.ipynb",
         "SINGLE": "02_Smoothing.ipynb",
         "DOUBLE": "02_Smoothing.ipynb",
         "__total_error": "03_Metrics.ipynb",
         "__absolute_error": "03_Metrics.ipynb",
         "__squared_error": "03_Metrics.ipynb",
         "__mean_error": "03_Metrics.ipynb",
         "__mean_absolute_error": "03_Metrics.ipynb",
         "__mean_squared_error": "03_Metrics.ipynb",
         "__root_mean_square_error": "03_Metrics.ipynb",
         "__mean_percentage_error": "03_Metrics.ipynb",
         "__mean_absolute_percent_error": "03_Metrics.ipynb",
         "SUMMARY": "03_Metrics.ipynb",
         "minimize_objective": "04_Optimizers.ipynb",
         "minimize_simple_smoothing": "04_Optimizers.ipynb",
         "minimize_double_smoothing": "04_Optimizers.ipynb"}

modules = ["simulate_data.py",
           "moving_average.py",
           "smoothing.py",
           "metrics.py",
           "optimizers.py"]

doc_url = "https://kikejimenez.github.io/time_series_model_basics/"

git_url = "https://github.com/kikejimenez/time_series_model_basics/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_Optimizers.ipynb (unless otherwise specified).

__all__ = ['minimize_objective', 'minimize_simple_smoothing', 'minimize_double_smoothing']

# Cell
from os import path
import requests
import numpy as np
import pandas as pd
from typing import List, Callable, Tuple
from skopt import gp_minimize
from time_series_model_basics import moving_average, smoothing, metrics, simulate_data
from functools import partial

# Cell
def minimize_objective(
    objective: Callable,
    param_names: List[str],
    param_ranges: List[Tuple[float, float]],
    *,
    ts: np.array,
    score_f: Callable = metrics.__mean_absolute_error,
) -> Tuple[float, float]:

    def smoothing_partial(*args: List[float]) -> float:

        forecast = smoothing.__single(
            ts=ts,
            alpha=alpha[0],
        )
        val = score_f(ts=ts, f=forecast)
        return val

    res = gp_minimize(
        smoothing_partial,
        [(0.0, 1), (0.0, 1)],
        n_calls=20,
    )
    print("x*=%.2f f(x*)=%.2f" % (res.x[0], res.fun))
    return res.x[0]

# Cell
def minimize_simple_smoothing(
    ts: np.array,
    score_f=metrics.__mean_absolute_error,
) -> float:

    def smoothing_partial(alpha: List[float]) -> float:
        forecast = smoothing.__single(
            ts=ts,
            alpha=alpha[0],
        )
        val = score_f(ts=ts, f=forecast)
        return val

    res = gp_minimize(
        smoothing_partial,
        [(0.0, 1)],
        n_calls=20,
    )
    print("x*=%.2f f(x*)=%.2f" % (res.x[0], res.fun))
    return res.x[0]

# Cell
def minimize_double_smoothing(
    ts: np.array,
    score_f=metrics.__mean_absolute_error,
) -> float:

    def smoothing_partial(args: List[float]) -> float:

        forecast = smoothing.__double(
            ts=ts,
            alpha=args[0],
            beta=args[1],
        )
        val = score_f(ts=ts, f=forecast)
        return val

    res = gp_minimize(
        smoothing_partial,
        [(0.0, 1), (0.0, 1)],
        n_calls=20,
    )
    print("x*=%.2f, y*=%.2f, f(x*,y*)=%.2f" % (
        res.x[0],
        res.x[1],
        res.fun,
    ))
    return res.x


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_MovingAverage.ipynb (unless otherwise specified).

__all__ = ['simple', 'weighted', '__moving_average', 'SMA', 'WMA']

# Cell
import numpy as np
from numba import jit
import pandas as pd

from time_series_model_basics import simulate_data

# Cell
@jit(nopython=True)
def simple(
    ts,
    n=1,
):

    l = len(ts)
    if n > l:
        raise Exception('n should be less than the length of the time-series')

    f = np.zeros((l,))
    f[:n] = np.full(n, np.nan)

    for i in range(l - n + 1):
        f[n + i] = np.mean(ts[i:n + i])
    return f

# Cell
#@jit(nopython=True)
def weighted(
    ts,
    ws,
):
    if type(ws) is list:
        ws = np.array(ws, dtype=float)

    l = len(ts)
    n = len(ws)
    w = np.sum(ws)

    if n >= l:
        raise Exception('The length of ws should be less than the length of ts')

    f = np.full((l,), np.nan)
    #f[:n] = np.full(n, np.nan)

    for i in range(l - n ):
        f[n + i] = (1 / w) * np.dot(ws, ts[i:n + i])
    return f

# Cell
def __moving_average(
    *args,
    kind='simple',
    df=None,
    ts_col=None,
    **kwargs,
):
    """
    Forecasts  with a Moving Average

    -----
    Parameters
    -----

    kind : str , default 'simple'. It can be:

         'simple' : Simple Moving Average.
         'weighted' : Weighted Moving Average.
    df : DataFrame, default None. If df is None a dataframe
         with simulated data is generated.

    ts_col : str, default None. The column name with the
         time series. Set to 'time_series' when df is None.
    *args : int.  Each value represent a Moving
           Average Forecast and its corresponding
           window size
    **kwargs : passed to __pandas_time_series

    -------
    Returns
    -------

    tuple: First element is the Pandas Dataframe and the second
           is ploty's figure object
    """

    if len(args) == 0:
        raise Exception("list in *args should include at least one value")

    if kind not in ('simple', 'weighted'):
        raise Exception("Only supported methods are: 'simple', 'weighted'")

    if df is None:
        df = simulate_data.pandas_time_series(**kwargs)
        ts_col = 'time_series'

    if ts_col is None:
        raise Exception('When df is not None, ts_col should be specified')

    ts = df[ts_col].to_numpy()
    cols = [ts_col]
    for a in args:
        col = f'{kind}_{a}'
        cols.append(col)
        if kind == 'simple':
            df[col] = simple(ts=ts, n=a)
        elif kind == 'weighted':
            df[col] = weighted(ts=ts, ws=a)

    fig = df[cols].plot(
        backend='plotly',
        title=f'Moving Average: {kind}',
    )

    fig.update_layout(template="plotly_dark",)

    return df, fig

# Cell
def SMA(*args, df=None, ts_col=None, **kwargs):
    """
    Forecasts  with Simple Moving Average

    -----
    Parameters
    -----

    df : DataFrame, default None. If df is None a dataframe
         with simulated data is generated.
    ts_col : str, default None. The column name with the
         time series. Set to 'time_series' when df is None.
    *args : int.  Each value represent a Moving
           Average Forecast and its corresponding
           window size
    **kwargs : passed to __pandas_time_series

    -------
    Returns
    -------

    tuple: First element is the Pandas Dataframe and the second
           is ploty's fig object
    """

    return __moving_average(*args,
                            kind='simple',
                            df=df,
                            ts_col=ts_col,
                            **kwargs)

# Cell
def WMA(*args, df=None, ts_col='time_series', **kwargs):
    """
    Forecasts with Weighted Moving Average

    -----
    Parameters
    -----
    df : DataFrame, default None. If df is None a dataframe
         with simulated data is generated.
    ts_col : str, default 'time_series'. The column name with the
         time series. Ignored if df is None.
    *args : int.  Each value represent a Moving
           Average Forecast and its corresponding
           window size
    **kwargs : passed to __pandas_time_series
    -------
    Returns
    -------
    tuple: First element is the Pandas Dataframe and the second
           is ploty's fig object
    """

    return __moving_average(*args,
                            kind='weighted',
                            df=df,
                            ts_col=ts_col,
                            **kwargs)